/* Everything about the rules for NAT. */
#define __NO_VERSION__
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/bitops.h>

#include "NAT/ip_nat_rule.h"
#include "NAT/ip_nat_protocol.h"
#include "common/listhelp.h"

static LIST_HEAD(nat_rules);
static LIST_HEAD(mapping_types);

static inline int lookup_cmp(const struct ip_nat_rule *ptr,
			     const struct ip_conntrack_tuple *tuple)
{
	size_t i;
	unsigned long cmp = 0;

	/* Apply mask.  Should unroll nicely */
	for (i = 0; i < sizeof(*tuple)/sizeof(unsigned long); i++) {
		cmp |= (((unsigned long *)tuple)[i]
			& ((unsigned long *)&ptr->mask)[i])
			^ ((unsigned long *)&ptr->match)[i];
	}

	return (cmp == 0);
}

struct ip_nat_rule *
find_rule(const struct ip_conntrack_tuple *tuple)
{
	return LIST_FIND(&nat_rules, lookup_cmp, struct ip_nat_rule *, tuple);
}

/* NAT rule tables are ordered, similar to routing tables: most
 * specific to least specific.  I've tried (and maybe failed) to be
 * intuitive here.  Higher the better.  This is a partial ordering.  */
static u_int32_t
precedence(const struct ip_conntrack_tuple *mask)
{
	u_int32_t prec = 0;

	/* We build this up most important to least. */

	/* Source IP bits + destination IP bits: 7 bits */
	prec |= hweight32(mask->src.ip) + hweight32(mask->dst.ip);

	/* Do we care about protocol?: 1 bit */
	prec <<= 1;
	prec |= mask->dst.protonum ? 1 : 0;

	/* Per-protocol information: 6 bits */
	prec <<= 6;
	prec |= hweight16(mask->src.u.all) + hweight16(mask->dst.u.all);

	return prec;
}

#ifndef ALWAYS_IN_INTERRUPT
/* Must NOT have lock. */
static void load_mapping_type(const char *name)
{
	char modulename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN+sizeof("ip_nat_map_")];

	strcpy(modulename, "ip_nat_map_");
	strcat(modulename, name);
	DEBUGP("ip_nat: requesting module `%s'\n", modulename);
	request_module(modulename);
}
#endif

int ip_nat_register_mapping_type(struct ip_nat_mapping_type *me)
{
	unsigned long flags;
	int ret;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	if (list_named_find(&mapping_types, me->name))
		ret = -EEXIST;
	else {
		MOD_INC_USE_COUNT;
		list_prepend(&mapping_types, me);
		ret = 0;
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	return ret;
}

void ip_nat_unregister_mapping_type(struct ip_nat_mapping_type *me)
{
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	LIST_DELETE(&mapping_types, me);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	MOD_DEC_USE_COUNT;
}

/* Return 0 or -errno */
static int rule_user_to_rule(const struct ip_nat_rule_user *user,
			     struct ip_nat_rule **rule,
			     struct ip_nat_mapping_type *mtype,
			     void *moptions,
			     size_t extra_range)
{
	if (!user->mask.dst.protonum
	    && (user->mask.src.u.all || user->mask.dst.u.all
		|| (user->range.flags & IP_NAT_RANGE_PROTO_SPECIFIED))) {
		DEBUGP("rule_user_to_rule: bad proto specifics\n");
		return -EINVAL;
	}

	*rule = kmalloc(sizeof(struct ip_nat_rule) 
			+ sizeof(struct ip_nat_range)*extra_range,
#ifdef ALWAYS_IN_INTERRUPT
			GFP_ATOMIC
#else
			GFP_KERNEL
#endif
		);
	if (!*rule)
		return -ENOMEM;

	**rule = ((struct ip_nat_rule){ { NULL, NULL },
					user->match, user->mask,
					user->maniptype,
					precedence(&user->mask),
					mtype, moptions,
					{ 1+extra_range, { user->range } } });
	
	/* Make sure padding is zero. */
	(*rule)->match.src.pad = (*rule)->mask.src.pad = 0;

	if (mtype && !mtype->check(*rule)) {
		kfree(*rule);
		return -EINVAL;
	}
	return 0;
}

/* We create an expectation by adding a rule with a special mapping type;
   rule has a struct timer_list on the end. */
static unsigned int 
expect_setup(struct ip_conntrack *conntrack,
	     enum ip_nat_manip_type maniptype,
	     struct ip_nat_multi_range *mr,
	     const struct net_device *device,
	     void *moptions)
{
	unsigned int ret = ip_nat_setup_info(conntrack, maniptype, mr, device);
	struct ip_nat_rule *rule = find_rule(&conntrack->tuplehash[0].tuple);
	struct timer_list *timer = (void *)(rule + 1);

	/* Delete the rule which created us. */
	LIST_DELETE(&nat_rules, rule);
	del_timer(timer);
	kfree(rule);
	return ret;
}

static struct ip_nat_mapping_type mapping_expect
= { { NULL, NULL }, "expected", NULL, NULL, expect_setup };

static void timeout_rule(unsigned long ul_rule)
{
	struct ip_nat_rule *rule = (struct ip_nat_rule *)ul_rule;
	unsigned long flags;

	DEBUGP("timeout_rule: giving up on expected %p.\n", rule);
	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	LIST_DELETE(&nat_rules, rule);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	kfree(rule);
}

/* Returns TRUE if it successfully creates expectation. */
int ip_nat_expect(const struct ip_conntrack_tuple *expect,
		  const struct ip_conntrack_manip *manip,
		  enum ip_nat_manip_type maniptype,
		  unsigned long timeout)
{
	struct {
		struct ip_nat_rule rule;
		struct timer_list timer;
	} *expect_rule;

	expect_rule = kmalloc(sizeof(*expect_rule), GFP_ATOMIC);
	if (!expect_rule) {
		DEBUGP("ip_nat_expect: oom!\n");
		return 0;
	}

	expect_rule->rule = ((struct ip_nat_rule) 
			     { { NULL, NULL }, *expect,
			       { { 0xFFFFFFFF, { 0 }, 0 },
				 { 0xFFFFFFFF, { 0xFFFF }, 0xFF } },
			       maniptype, 0,
			       &mapping_expect, NULL,
			       { 1, 
				 { { IP_NAT_RANGE_MAP_IPS
				     | IP_NAT_RANGE_PROTO_SPECIFIED,
				     manip->ip, manip->ip,
				     manip->u, manip->u } }
			       } });
	init_timer(&expect_rule->timer);
	expect_rule->timer.data = (unsigned long)expect_rule;
	expect_rule->timer.function = timeout_rule;
	expect_rule->timer.expires = jiffies + timeout;
	add_timer(&expect_rule->timer);

#if 0
	DEBUGP("ip_nat_expect: %u.%u.%u.%u:%u->%u.%u.%u.%u:%u proto %u.\n"
	       "ip_nat_mask:   %u.%u.%u.%u:%u->%u.%u.%u.%u:%u proto %u.\n",
	       IP_PARTS(expect_rule->rule.match.src.ip), 
	       ntohs(expect_rule->rule.match.src.u.all),
	       IP_PARTS(expect_rule->rule.match.dst.ip), 
	       ntohs(expect_rule->rule.match.dst.u.all),
	       expect_rule->rule.match.dst.protonum,
	       IP_PARTS(expect_rule->rule.mask.src.ip), 
	       ntohs(expect_rule->rule.mask.src.u.all),
	       IP_PARTS(expect_rule->rule.mask.dst.ip), 
	       ntohs(expect_rule->rule.mask.dst.u.all),
	       expect_rule->rule.mask.dst.protonum);
#endif

	/* Expect rules must take precendence. */
	list_prepend(&nat_rules, &expect_rule->rule);
	return 1;
}

static inline int prec_cmpfn(struct ip_nat_rule *new, struct ip_nat_rule *i)
{
	return (new->precedence > i->precedence);
}

static int
rule_usercmpfn(const struct ip_nat_rule *i, 
	       const struct ip_nat_rule_user *find)
{
	/* Compare match and mask */
	return memcmp(&find->match, &i->match, 
		      sizeof(struct ip_conntrack_tuple)) == 0
		&& memcmp(&find->mask, &i->mask, 
			  sizeof(struct ip_conntrack_tuple)) == 0;
}

static int insert_rule(struct ip_nat_rule_user *ruleuser,
		       struct ip_nat_mapping_type *mtype,
		       void *moptions)
{
	struct ip_nat_rule *match, *newrule;

	match = LIST_FIND(&nat_rules, rule_usercmpfn, 
			  struct ip_nat_rule *, ruleuser);
	if (match) {
		int ret;

		/* Manipulation type and mapping type must be same;
		 * don't try to merge mappings with options. */
		if (match->maniptype != ruleuser->maniptype
		    || match->mtype != mtype
		    || moptions) {
			DEBUGP("insert_rule: %u vs %u, mtype `%s' vs `%s'\n",
			       match->maniptype, ruleuser->maniptype,
			       match->mtype ? match->mtype->name : "",
			       mtype->name);
			return -ENOTEMPTY;
		}

		ret = rule_user_to_rule(ruleuser, &newrule, mtype, moptions,
					match->mr.rangesize);
		if (ret != 0)
			return ret;

		/* Copy old range across. */
		memcpy(&newrule->mr.range[1], match->mr.range, 
		       sizeof(struct ip_nat_range) 
		       * (newrule->mr.rangesize - 1));
		LIST_DELETE(&nat_rules, match);
		kfree(match);
	} else {
		int ret;

		ret = rule_user_to_rule(ruleuser, &newrule, mtype, moptions,0);
		if (ret != 0)
			return ret;

		if (mtype && mtype->me)
			__MOD_INC_USE_COUNT(mtype->me);
	}

	LIST_INSERT(&nat_rules, newrule, prec_cmpfn);
	return 0;
}

/* This can't be used to delete expected rules, since the precedence
   calculated here won't match the 0 used in ip_nat_expect() above. */
static int delete_rule(const struct ip_nat_rule_user *ruleuser,
		       struct ip_nat_mapping_type *mtype,
		       void *moptions)
{
	struct ip_nat_rule *match;
	unsigned int rangenum;

	match = LIST_FIND(&nat_rules, rule_usercmpfn, 
			  struct ip_nat_rule *, ruleuser);
	if (!match)
		return -ENOENT;

	/* maniptype and mapping type must be same. */
	if (match->maniptype != ruleuser->maniptype
	    || match->mtype != mtype) {
		DEBUGP("delete_rule: %u vs %u, mtype `%s' vs `%s'\n",
		       match->maniptype, ruleuser->maniptype,
		       match->mtype ? match->mtype->name : "",
		       ruleuser->typename);
		return -ENOENT;
	}

	/* Find range in rule. */
	for (rangenum = 0; rangenum < match->mr.rangesize; rangenum++) {
		if (memcmp(&ruleuser->range, &match->mr.range[rangenum],
			   sizeof(ruleuser->range)) == 0) {
			break;
		}
	}
	if (rangenum == match->mr.rangesize) {
		DEBUGP("delete_rule: Found rule, but not matching range\n");
		return -ENOENT;
	}

	if (match->mr.rangesize == 1) {
		IP_CONNTRACK_ASSERT(rangenum == 0);

		LIST_DELETE(&nat_rules, match);
		if (match->mtype && match->mtype->me)
			__MOD_DEC_USE_COUNT(match->mtype->me);
		if (match->moptions)
			kfree(match->moptions);
		kfree(match);
	} else {
		/* Move down rules after match. */
		memmove(&match->mr.range[rangenum],
			&match->mr.range[rangenum+1],
			(match->mr.rangesize - rangenum - 1)
			* sizeof(struct ip_nat_range));
		match->mr.rangesize--;
	}

	return 0;
}

static int do_ip_nat_ctl(
#ifdef HAVE_NF_GETSOCKOPT
	struct sock *sk,
#endif
	int cmd, void *user, unsigned int len)
{
	int ret;
	struct ip_nat_rule_user ruleuser;

	if (len < sizeof(struct ip_nat_rule_user)) {
		DEBUGP("IP_NAT_SO_INSERT/DELETE: len %u < %u\n",
		       len, sizeof(struct ip_nat_rule_user));
		ret = -EINVAL;
	} else if (copy_from_user(&ruleuser, user, 
				  sizeof(struct ip_nat_rule_user))) {
		ret = -EFAULT;
	} else {
		struct ip_nat_mapping_type *mtype = NULL;
		unsigned long flags;
		void *moptions = NULL;

		/* Must be NUL terminated for sanity */
		ruleuser.typename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN-1] = '\0';

		LOCK_IRQSAVE(&ip_conntrack_lck, flags);
		if (ruleuser.typename[0]) {
			mtype = list_named_find(&mapping_types, 
						ruleuser.typename);

			if (!mtype) {
#ifndef ALWAYS_IN_INTERRUPT
				UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
				load_mapping_type(ruleuser.typename);
				LOCK_IRQSAVE(&ip_conntrack_lck, flags);
				mtype = list_named_find(&mapping_types, 
							ruleuser.typename);
#endif
				if (!mtype) {
					DEBUGP("rule_user: no mapping `%s'\n",
					       ruleuser.typename);
					ret = -EINVAL;
					goto out;
				}
			}
			if (len - sizeof(struct ip_nat_rule_user)
			    != mtype->optionlen) {
				DEBUGP("IP_NAT_INSERT/DELETE: len %u != %u\n",
				       len - sizeof(struct ip_nat_rule_user),
				       mtype->optionlen);
				ret = -EINVAL;
				goto out;
			}
			if (mtype->optionlen) {
				moptions = kmalloc(mtype->optionlen,
#ifdef ALWAYS_IN_INTERRUPT
						   GFP_ATOMIC
#else
						   GFP_KERNEL
#endif
					);
				if (!moptions) {
					ret = -ENOMEM;
					goto out;
				}
				if (copy_from_user(moptions, 
						   user 
						   + sizeof(struct
							    ip_nat_rule_user), 
						   mtype->optionlen) != 0) {
					ret = -EFAULT;
					goto out;
				}
			}
		}

		if (cmd == IP_NAT_SO_DELETE) {
			ret = delete_rule(&ruleuser, mtype, moptions);
			if (moptions)
				kfree(moptions);
		} else {
			IP_CONNTRACK_ASSERT(cmd == IP_NAT_SO_INSERT);
			ret = insert_rule(&ruleuser, mtype, moptions);
			if (ret != 0 && moptions)
				kfree(moptions);
		}
	out:
		UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	}

	return ret;
}

#ifndef HAVE_NF_GETSOCKOPT
static struct nf_setsockopt_ops ip_nat_sockopts
= { { NULL, NULL }, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX, do_ip_nat_ctl };
#else
static struct nf_sockopt_ops ip_nat_sockopts
= { { NULL, NULL }, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX+1, do_ip_nat_ctl,
    0, 0, NULL };
#endif

#ifdef CONFIG_PROC_FS
static unsigned int getmask(u_int32_t mask)
{
	int i;

	for (i = 0; mask & 0x80000000; i++) 
		mask <<= 1;

	return i;
}

static unsigned int
dump_rule(char *buffer, const struct ip_nat_rule *rule, unsigned int i)
{
	unsigned int len = 0;
	struct ip_nat_protocol *proto
		= find_nat_proto(rule->match.dst.protonum);

	len += sprintf(buffer + len, "%s [%s] ", 
		       rule->mtype ? rule->mtype->name : "generic",
		       rule->maniptype == IP_NAT_MANIP_SRC ? "SRC" : "DST");
	len += sprintf(buffer + len,
		       "%u.%u.%u.%u/%u->%u.%u.%u.%u/%u ",
		       IP_PARTS(rule->match.src.ip),
		       getmask(ntohl(rule->mask.src.ip)),
		       IP_PARTS(rule->match.dst.ip),
		       getmask(ntohl(rule->mask.dst.ip)));
	if (rule->mask.dst.protonum) {
		len += sprintf(buffer + len, "proto=%u ", 
			       rule->match.dst.protonum);
		len += proto->print(buffer + len, &rule->match, &rule->mask);
	}
	/* Print out range being mapped to. */
	len += sprintf(buffer + len, "TO: ");
	if (rule->mr.range[i].flags & IP_NAT_RANGE_MAP_IPS) {
		len += sprintf(buffer + len, "%u.%u.%u.%u",
			       IP_PARTS(rule->mr.range[i].min_ip));
		if (rule->mr.range[i].max_ip != rule->mr.range[i].min_ip)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS(rule->mr.range[i].max_ip));
		len += sprintf(buffer + len, " ");
	}
	if (rule->mr.range[i].flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		len += proto->print_range(buffer + len, &rule->mr.range[i]);

	len += sprintf(buffer + len, "\n");
	return len;
}

static int 
get_rules(char *buffer, char **start, off_t offset, int length, int dummy)
{
	off_t upto = 0;
	unsigned long flags;
	int len = 0,last_len = 0;
	struct list_head *i;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	for (i = nat_rules.next; i != &nat_rules; i = i->next) {
		struct ip_nat_rule *rule = (struct ip_nat_rule *)i;
		unsigned int j;

		for (j = 0; j < rule->mr.rangesize; j++, upto++) {
			if (upto < offset) continue;

			last_len = len;
			len += dump_rule(buffer+len, rule, j);
			if (len > length) {
				len = last_len;
				goto out;
			}
		}
	}

 out:
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}

struct proc_dir_entry ip_nat_rule_procinfo = {
	0, 12, "ip_nat_rules", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, get_rules
};
#endif

int ip_nat_rule_init(void)
{
	int ret;

	ret = nf_register_sockopt(&ip_nat_sockopts);
	if (ret == 0) {
#ifdef CONFIG_PROC_FS
		proc_net_register(&ip_nat_rule_procinfo);
#endif
		list_prepend(&mapping_types, &mapping_expect);
	}
	return ret;
}

void ip_nat_rule_cleanup(void)
{
	nf_unregister_sockopt(&ip_nat_sockopts);

	/* Delete all the rules. */
	while (!list_empty(&nat_rules)) {
		struct ip_nat_rule *rule = (void *)nat_rules.next;

		/* Careful: deactivate timers on expect rules! */
		if (rule->mtype == &mapping_expect) {
			struct timer_list *timer = (void *)(rule + 1);
			del_timer(timer);
		}
		LIST_DELETE(&nat_rules, rule);
		kfree(rule);
	}

	LIST_DELETE(&mapping_types, &mapping_expect);
#ifdef CONFIG_PROC_FS
	proc_net_unregister(ip_nat_rule_procinfo.low_ino);
#endif
}
