/* Connection state tracking for netfilter.  This is separated from,
   but required by, the NAT layer; it can also be used by an iptables
   extention. */

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/stddef.h>

#include "conntrack/struct_reserve.h"
#include "conntrack/ip_conntrack.h"
#include "conntrack/ip_conntrack_protocol.h"
#include "conntrack/ip_conntrack_helper.h"
#include "conntrack/ip_conntrack_core.h"
#include "common/listhelp.h"

DECLARE_LOCK(ip_conntrack_lck);

#ifdef DEBUG_CONNTRACK
static void dump_tuple(const struct ip_conntrack_tuple *tuple) __attribute__((unused));

static void dump_tuple(const struct ip_conntrack_tuple *tuple)
{
	printk("tuple %p: %u %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
	       tuple, tuple->dst.protonum,
	       IP_PARTS(tuple->src.ip), ntohs(tuple->src.u.tcp.port),
	       IP_PARTS(tuple->dst.ip), ntohs(tuple->dst.u.tcp.port));
}
#endif

void (*ip_conntrack_destroyed)(struct ip_conntrack *conntrack) = NULL;
LIST_HEAD(expect_list);
LIST_HEAD(protocol_list);
static LIST_HEAD(helpers);
struct list_head ip_conntrack_hash[IP_CONNTRACK_HTABLE_SIZE];
static unsigned int ip_conntrack_gen = NF_REASON_MIN_RESERVED_FOR_CONNTRACK+1;

extern struct ip_conntrack_protocol ip_conntrack_generic_protocol;

DECLARE_STRUCT_RESERVABLE(ip_conntrack);

static inline int cmpfn(const struct ip_conntrack_protocol *curr,
			u_int8_t protocol)
{
	return protocol == curr->proto;
}

struct ip_conntrack_protocol *find_proto(u_int8_t protocol)
{
	struct ip_conntrack_protocol *p
		= LIST_FIND(&protocol_list, cmpfn, 
			    struct ip_conntrack_protocol *, protocol);

	if (!p)
		p = &ip_conntrack_generic_protocol;

	return p;
}

static inline u_int32_t 
hash_conntrack(const struct ip_conntrack_tuple *tuple)
{
#if 0
	dump_tuple(tuple);
#endif
#ifdef DEBUG_CONNTRACK
	if (tuple->src.pad)
		DEBUGP("Tuple %p has non-zero padding.\n", tuple);
#endif
	/* ntohl because more differences in low bits. */
	/* To ensure that halves of the same connection don't hash
	   clash, we add the source per-proto again. */
	return (ntohl(tuple->src.ip + tuple->dst.ip
		     + tuple->src.u.all + tuple->dst.u.all
		     + tuple->dst.protonum)
		+ ntohs(tuple->src.u.all))
		% IP_CONNTRACK_HTABLE_SIZE;
}

inline int 
get_tuple(const struct iphdr *iph, size_t len,
	  struct ip_conntrack_tuple *tuple,
	  struct ip_conntrack_protocol *protocol)
{
	int ret;

	/* FIXME:  Be clever.  --RR */
	if (ntohs(iph->frag_off) & IP_OFFSET) {
		static int warned = 0;
		if (!warned)
			printk("ip_conntrack_core: Frag of proto %u.\n", 
			       iph->protocol);
		warned = 1;
		return 0;
	}
	/* Guarentee 8 protocol bytes: if more wanted, use len param */
	else if (iph->ihl * 4 + 8 > len)
		return 0;

	tuple->src.ip = iph->saddr;
	tuple->src.pad = 0;
	tuple->dst.ip = iph->daddr;
	tuple->dst.protonum = iph->protocol;

	ret = protocol->pkt_to_tuple((u_int32_t *)iph + iph->ihl,
				     len - 4*iph->ihl,
				     tuple);
	return ret;
}

/* Simple conntrack table: indexed by status and direction. */
static struct 
{
	/* new status. */
	enum ip_conntrack_status status;

	/* conntrack info for skb ptr. */
	enum ip_conntrack_info conntrackinfo;
} ct_state[3][2] = {
	/* IPS_AWAITING_REPLY */
	{ { IPS_AWAITING_REPLY, IP_CT_NEW },
	  { IPS_BOTH_WAYS, IP_CT_ESTABLISHED } },

	/* IPS_BOTH_WAYS */
	{ { IPS_BOTH_WAYS, IP_CT_ESTABLISHED },
	  { IPS_BOTH_WAYS, IP_CT_ESTABLISHED } },

	/* IPS_EXPECTED_AWAITING_REPLY */
	{ { IPS_EXPECTED_AWAITING_REPLY, IP_CT_RELATED },
	  { IPS_BOTH_WAYS, IP_CT_ESTABLISHED } }
};

static inline int 
conntrack_cmp(struct ip_conntrack_tuple_hash *i, 
	      const struct ip_conntrack_tuple *tuple)
{
	return !i->dead && (memcmp(tuple, &i->tuple, sizeof(*tuple)) == 0);
}

/* find a connection corresponding to a tuple. */
static struct ip_conntrack_tuple_hash *
find_conntrack(const struct ip_conntrack_tuple *tuple)
{
	return LIST_FIND(&ip_conntrack_hash[hash_conntrack(tuple)],
			 conntrack_cmp,
			 struct ip_conntrack_tuple_hash *,
			 tuple);
}

static int
invert_tuple(struct ip_conntrack_tuple *inverse,
	     const struct ip_conntrack_tuple *orig,
	     const struct ip_conntrack_protocol *protocol)
{
	inverse->src.ip = orig->dst.ip;
	inverse->src.pad = 0;
	inverse->dst.ip = orig->src.ip;
	inverse->dst.protonum = orig->dst.protonum;

	return protocol->invert_tuple(inverse, orig);
}

static void free_conntrack(struct ip_conntrack *conntrack)
{
	MUST_BE_LOCKED(&ip_conntrack_lck);
	if (conntrack->timeout.function) {
		del_timer(&conntrack->timeout);
	}
	/* If our expected is in the list, take it out. */
	if (list_inlist(&expect_list, &conntrack->expected)) {
		LIST_DELETE(&expect_list, &conntrack->expected);
	}
	if (ip_conntrack_destroyed) {
		ip_conntrack_destroyed(conntrack);
	}

	if (++ip_conntrack_gen == 0) {
		ip_conntrack_gen = NF_REASON_MIN_RESERVED_FOR_CONNTRACK + 1;
	}

	/* Remove from both hash lists */
	LIST_DELETE(&ip_conntrack_hash
		    [hash_conntrack(&conntrack->tuplehash[0].tuple)],
		    &conntrack->tuplehash[0]);
	LIST_DELETE(&ip_conntrack_hash
		    [hash_conntrack(&conntrack->tuplehash[1].tuple)],
		    &conntrack->tuplehash[1]);

	kfree(conntrack);
}

/* Delete a conntrack. */
void ip_conntrack_delete(struct ip_conntrack *conntrack)
{
	/* Grab before deleted. */
	int reply_dead = conntrack->tuplehash[1].dead;

#ifdef DEBUG_CONNTRACK
	if (conntrack->tuplehash[0].dead && reply_dead) {
		DEBUGP("Conntrack %p won't die!! ARGH DIE DIE DIE.\n",
		       conntrack);
		free_conntrack(conntrack);
		return;
	}
#endif

	/* Mark it dead in both directions. */
	if (!conntrack->tuplehash[0].dead) {
		ip_conntrack_delete_maybe(conntrack, 0);
	}

	if (!reply_dead) {
		ip_conntrack_delete_maybe(conntrack, 1);
	}
}

/* Expire in this direction.  If expired in both dirs, delete. */
void ip_conntrack_delete_maybe(struct ip_conntrack *conntrack, int direction)
{
	MUST_BE_LOCKED(&ip_conntrack_lck);
	IP_CONNTRACK_ASSERT(conntrack);
	IP_CONNTRACK_ASSERT(!conntrack->tuplehash[direction].dead);

	conntrack->tuplehash[direction].dead = 1;
	ip_conntrack_put(conntrack);
}

void ip_conntrack_get(struct ip_conntrack *ct)
{
	MUST_BE_LOCKED(&ip_conntrack_lck);
	ct->use++;
}

void ip_conntrack_put(struct ip_conntrack *ct)
{
	MUST_BE_LOCKED(&ip_conntrack_lck);
	if (--ct->use == 0) {
		IP_CONNTRACK_ASSERT(ct->tuplehash[0].dead
				    && ct->tuplehash[1].dead);
		free_conntrack(ct);
	}
}

#ifdef DEBUG_CONNTRACK
static int 
cmp_cthash(struct ip_conntrack_tuple_hash *i, struct ip_conntrack *conntrack)
{
	return i->ctrack == conntrack;
}
#endif /* DEBUG_CONNTRACK */

static void ip_conntrack_delete_timeout(unsigned long ul_conntrack)
{
	unsigned long flags;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
#ifdef DEBUG_CONNTRACK
	{
		unsigned int i;

		for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
			if (LIST_FIND(&ip_conntrack_hash[i], cmp_cthash,
				      struct ip_conntrack_tuple_hash *,
				      (void *)ul_conntrack))
				goto out;
		}
		DEBUGP("Delete timeout for %p, which isn't hashed.\n",
		       (void *)ul_conntrack);
		UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
		return;
	}
 out:
#endif
	ip_conntrack_delete((struct ip_conntrack *)ul_conntrack);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
}

/* Compare all but src per-proto part. */
static int expect_cmp(const struct ip_conntrack_expect *i,
		      const struct ip_conntrack_tuple *tuple)
{
	return (tuple->src.ip == i->tuple.src.ip
		&& tuple->dst.ip == i->tuple.dst.ip
		&& tuple->dst.u.all == i->tuple.dst.u.all
		&& tuple->dst.protonum == i->tuple.dst.protonum);
}

/* Returns TRUE if it dealt with the skb. */
int icmp_error_track(const struct iphdr *iph, size_t len,
		     enum ip_conntrack_info *info, 
		     struct ip_conntrack_tuple_hash **h)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_conntrack_tuple innertuple, origtuple;
	struct iphdr *inner = (struct iphdr *)(hdr + 1);
	size_t datalen = len - iph->ihl*4 - sizeof(*hdr);

	struct ip_conntrack_protocol *innerproto;

	if (len < iph->ihl * 4 + sizeof(struct icmphdr)) {
		DEBUGP("icmp_error_track: too short\n");
		goto inval;
	}

	if (hdr->type != ICMP_DEST_UNREACH
	    && hdr->type != ICMP_SOURCE_QUENCH
	    && hdr->type != ICMP_TIME_EXCEEDED
	    && hdr->type != ICMP_PARAMETERPROB
	    && hdr->type != ICMP_REDIRECT)
		return 0;

	/* Ignore it if the checksum's bogus. */
	if (ip_compute_csum((unsigned char *)hdr, sizeof(*hdr) + datalen)) {
		DEBUGP("icmp_error_track: bad csum\n");
		goto inval;
	}

	innerproto = find_proto(inner->protocol);
	/* Are they talking about one of our connections? */
	if (inner->ihl * 4 + 8 > datalen
	    || !get_tuple(inner, datalen, &origtuple, innerproto)) {
		DEBUGP("icmp_error: ! get_tuple p=%u (%u*4+%u dlen=%u)\n",
		       inner->protocol, inner->ihl, 8,
		       datalen);
		goto inval;
	}

	/* Ordinarily, we'd expect the inverted tupleproto, but it's
	   been preserved inside the ICMP. */
	if (!invert_tuple(&innertuple, &origtuple, innerproto)) {
		DEBUGP("icmp_error_track: Can't invert tuple\n");
		goto inval;
	}
	*h = find_conntrack(&innertuple);
	if (!*h) {
		DEBUGP("icmp_error_track: no match\n");
		goto inval;
	}

	*info = IP_CT_RELATED;
	return 1;

 inval:
	*info = IP_CT_INVALID;
	return 1;
}

static int helper_cmp(const struct ip_conntrack_helper *i,
		      u_int16_t protonum, u_int16_t protocol_dst)
{
	return (i->protocol == protonum
		&& i->protocol_dst == protocol_dst);
}

/* Allocate a new conntrack. */
static enum ip_conntrack_info 
init_conntrack(const struct ip_conntrack_tuple *tuple, 
	       struct ip_conntrack_tuple_hash **h,
	       struct ip_conntrack_protocol *protocol,
	       struct iphdr *iph,
	       size_t len)
{
	struct ip_conntrack *conntrack;
	struct ip_conntrack_tuple repl_tuple;
	size_t hash, repl_hash;
	struct ip_conntrack_expect *expected;
	enum ip_conntrack_info ret;

	if (!invert_tuple(&repl_tuple, tuple, protocol)) {	
		DEBUGP("Can't invert tuple.\n");
		return IP_CT_INVALID;
	}

	conntrack = STRUCT_RESERVE_ALLOCATE(ip_conntrack, GFP_ATOMIC);
	if (!conntrack) {
		DEBUGP("Can't allocate conntrack.\n");
		return IP_CT_INVALID;
	}

	hash = hash_conntrack(tuple);
	repl_hash = hash_conntrack(&repl_tuple);

	*conntrack = ((struct ip_conntrack)
		      {{ {{ NULL, NULL }, *tuple, 0, conntrack,
			  {offsetof(struct ip_conntrack_tuple_hash, offsets),
			   offsetof(struct ip_conntrack_tuple_hash, offsets)+1,
			   offsetof(struct ip_conntrack_tuple_hash, offsets)+2,
			   offsetof(struct ip_conntrack_tuple_hash, offsets)+3
			  } },
			 {{ NULL, NULL }, repl_tuple, 0, conntrack,
			  {offsetof(struct ip_conntrack_tuple_hash, offsets),
			   offsetof(struct ip_conntrack_tuple_hash, offsets)+1,
			   offsetof(struct ip_conntrack_tuple_hash, offsets)+2,
			   offsetof(struct ip_conntrack_tuple_hash, offsets)+3
			  } },
		      },
		       IPS_AWAITING_REPLY, 
		       2, /* Starts with 2 home references. */
		       { NULL, NULL, jiffies + protocol->expiry, 
			 (unsigned long)conntrack, NULL },
		       /* Expected */
		       {{ NULL, NULL }, { { 0, { 0 }, 0 }, { 0, { 0 }, 0 } } },
		       /* helper */ 
		       LIST_FIND(&helpers, helper_cmp,
				 struct ip_conntrack_helper *, 
				 tuple->dst.protonum, tuple->dst.u.all) });

	/* Sew in at head of hash list. */
	list_prepend(&ip_conntrack_hash[hash], &conntrack->tuplehash[0]);
	list_prepend(&ip_conntrack_hash[repl_hash], &conntrack->tuplehash[1]);

	init_timer(&conntrack->timeout);
	if (protocol->expiry) {
		conntrack->timeout.function = ip_conntrack_delete_timeout;
		add_timer(&conntrack->timeout);
	}

	*h = &conntrack->tuplehash[0];

	expected = LIST_FIND(&expect_list, expect_cmp, 
			     struct ip_conntrack_expect *, tuple);
	if (expected) {
		/* Welcome, Mr. Bond.  We've been expecting you... */
		conntrack->status = IPS_EXPECTED_AWAITING_REPLY;
		LIST_DELETE(&expect_list, expected);
		ret = IP_CT_RELATED;
	} else {
		ret = IP_CT_NEW;
	}
	protocol->new(conntrack, iph, len, &ret);

	return ret;
}

/* Returns info, and sets h.  If info == IP_CT_INVALID, h not set.
   Otherwise, tuple is filled in.  If h == NULL, means no existing
   connection found.  */
static enum ip_conntrack_info
find_existing_conntrack(const struct iphdr *iph,
			size_t len, 
			struct ip_conntrack_protocol *protocol,
			struct ip_conntrack_tuple_hash **h,
			struct ip_conntrack_tuple *tuple)
{
	enum ip_conntrack_info ret;

	if (iph->protocol == IPPROTO_ICMP
	    && icmp_error_track(iph, len, &ret, h))
		return ret;

	if (!get_tuple(iph, len, tuple, protocol))
		return IP_CT_INVALID;

	*h = find_conntrack(tuple);
	if (*h) {
		/* It exists! */
		ret = ct_state[(*h)->ctrack->status][DIRECTION(*h)]
			.conntrackinfo;
	}

	return ret;
}

static void
update_skb(struct sk_buff *skb,
	   enum ip_conntrack_info info,
	   struct ip_conntrack_tuple_hash *h)
{
	if (info == IP_CT_INVALID) {
		skb->nfmark = 0;
		skb->nfreason = NF_REASON_MIN_RESERVED_FOR_CONNTRACK;
	} else {
		IP_CONNTRACK_ASSERT(h);
		skb->nfmark = (unsigned long)&h->offsets[info];
		skb->nfreason = ip_conntrack_gen;
	}
}

static int
try_tuplehash(struct iphdr *iph, 
	      u_int8_t *offsetp, 
	      struct ip_conntrack_tuple_hash *th,
	      size_t len)
{
	/* Does this match? */
	struct ip_conntrack_tuple tuple;

	/* ICMP error? */
	if (iph->protocol == IPPROTO_ICMP
	    && offsetp == &th->offsets[IP_CT_RELATED])
		return 1;

	if (!get_tuple(iph, len, &tuple, find_proto(iph->protocol)))
		return 0;

	/* Tuples must match. */
	return (memcmp(&tuple, &th->tuple, sizeof(tuple)) == 0);
}

/* Return conntrack_info and tuple hash given skb. */
enum ip_conntrack_info
ip_conntrack_findinfo(struct sk_buff *skb, struct ip_conntrack_tuple_hash **h)
{
	enum ip_conntrack_info ret;
	u_int8_t *p = (u_int8_t *)skb->nfmark;

	MUST_BE_LOCKED(&ip_conntrack_lck);
	if (skb->nfreason <= NF_REASON_MIN_RESERVED_FOR_CONNTRACK) {
		/* Not a generation count: being used by someone else, 
		   or invalid. */
		DEBUGP("Skb %p INVALID reason = %u.\n", skb, skb->nfreason);
		return IP_CT_INVALID;
	} else if (skb->nfreason != ip_conntrack_gen) {
		/* Lookup again.  We do a complete search of the hash,
		   for now, because we want to find deleted ones as
		   well. */
		size_t i;

		for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
			struct list_head *l;
			for (l = ip_conntrack_hash[i].next;
			     l != &ip_conntrack_hash[i];
			     l = l->next) {
				*h = (struct ip_conntrack_tuple_hash *)l;
				if ((u_int8_t *)skb->nfmark
				    >= &(*h)->offsets[0]
				    && (u_int8_t *)skb->nfmark 
				    < &(*h)->offsets[IP_CT_NUMBER])
					goto out;
			}
		}
		/* No existing conntrack?  Mark it invalid. */
		DEBUGP("Skb %p now INVALID.\n", skb);
		update_skb(skb, IP_CT_INVALID, NULL);
		return IP_CT_INVALID;

	out:
		if (!try_tuplehash(skb->nh.iph, (u_int8_t *)skb->nfmark, 
				   *h, skb->len)) {
			DEBUGP("Skb %p now INVALID.\n", skb);
			update_skb(skb, IP_CT_INVALID, NULL);
			return IP_CT_INVALID;
		}
	}

	IP_CONNTRACK_ASSERT((size_t)(p - *p) % sizeof(void *) == 0);
	*h = (struct ip_conntrack_tuple_hash *)(p - *p);

	ret = *p - offsetof(struct ip_conntrack_tuple_hash,
			    offsets);
	IP_CONNTRACK_ASSERT(ret >= 0 && ret < IP_CT_NUMBER);
	IP_CONNTRACK_ASSERT(ret != IP_CT_NEW || DIRECTION(*h) == 0);

	update_skb(skb, ret, *h);
	return ret;
}

/* Netfilter hook itself. */
unsigned int ip_conntrack_in(unsigned int hooknum,
			     struct sk_buff **pskb,
			     const struct net_device *in,
			     const struct net_device *out)
{
	unsigned int ret = NF_ACCEPT;
	struct ip_conntrack_tuple_hash *h;
	struct iphdr *iph = (*pskb)->nh.iph;
	unsigned long flags;
	enum ip_conntrack_info info;
	struct ip_conntrack_protocol *proto;
	struct ip_conntrack_tuple tuple;

	/* FIXME: Do this right please. --RR */
	(*pskb)->nfcache |= NFC_UNKNOWN;

	/* Not for us, or previously seen (loopback)?  Ignore. */
	if ((*pskb)->nfreason)
		return NF_ACCEPT;

	/* Gather fragments. */	
	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
		*pskb = ip_defrag(*pskb);
		if (!*pskb) {
			DEBUGP("ip_defrag took it!\n");
			return NF_STOLEN;
		} else {
			DEBUGP("ip_defrag gave new packet!\n");
		}
		iph = (*pskb)->nh.iph;
		ip_send_check(iph);
		(*pskb)->nfcache |= NFC_ALTERED;
	}

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);

	proto = find_proto(iph->protocol);
	info = find_existing_conntrack(iph, (*pskb)->len, proto, &h, &tuple);
	if (info != IP_CT_INVALID) {
		if (h && info != IP_CT_RELATED) {
			int dir = DIRECTION(h);

			/* Update status. */
			h->ctrack->status
				= ct_state[h->ctrack->status][dir].status;

			ret = proto->established(h->ctrack, iph, 
						 (*pskb)->len, dir, &info);
			if (ret != NF_DROP
			    && info != IP_CT_INVALID 
			    && h->ctrack->helper) {
				ret = h->ctrack->helper->help(iph, 
							      (*pskb)->len, h,
							      &info);
			}
		} else if (!h) {
			/* Nothing wrong, just couldn't find packet */
			info = init_conntrack(&tuple, &h, proto, iph,
					      (*pskb)->len);
		}
	}

	update_skb(*pskb, info, h);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
	return ret;
}

static int 
conntrack_tuple_cmp(const struct ip_conntrack_tuple_hash *i,
		    const struct ip_conntrack_tuple *tuple,
		    const struct ip_conntrack *ignored_conntrack)
{
	return (!i->dead && i->ctrack != ignored_conntrack
		&& memcmp(tuple, &i->tuple, sizeof(*tuple)) == 0);
}

/* Find a connection corresponding to a tuple. */
struct ip_conntrack_tuple_hash *
ip_conntrack_find(const struct ip_conntrack_tuple *tuple,
		  const struct ip_conntrack *ignored_conntrack)
{
	return LIST_FIND(&ip_conntrack_hash[hash_conntrack(tuple)],
			 conntrack_tuple_cmp, 
			 struct ip_conntrack_tuple_hash *,
			 tuple, ignored_conntrack);
}

int invert_tuplepr(struct ip_conntrack_tuple *inverse,
		   const struct ip_conntrack_tuple *orig)
{
	return invert_tuple(inverse, orig, find_proto(orig->dst.protonum));
}

static void
add_hash(struct ip_conntrack_tuple_hash *hash)
{
	unsigned int index = hash_conntrack(&hash->tuple);
	list_prepend(&ip_conntrack_hash[index], hash);
}

/* Alter reply tuple. */
void ip_conntrack_alter_reply(struct ip_conntrack *conntrack, 
			      const struct ip_conntrack_tuple *newreply)
{
	MUST_BE_LOCKED(&ip_conntrack_lck);

	LIST_DELETE(&ip_conntrack_hash
		    [hash_conntrack(&conntrack->tuplehash[1].tuple)],
		    &conntrack->tuplehash[1]);
	conntrack->tuplehash[1].tuple = *newreply;
	add_hash(&conntrack->tuplehash[1]);
}

int ip_conntrack_helper_register(struct ip_conntrack_helper *me)
{
	unsigned long flags;
	int ret;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	if (LIST_FIND(&helpers, helper_cmp,
		      struct ip_conntrack_helper *, 
		      me->protocol, me->protocol_dst)) {
		ret = -EBUSY;
	} else {
		ret = 0;
		list_prepend(&helpers, me);
		MOD_INC_USE_COUNT;
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);

	return ret;
}

static int unhelp(struct ip_conntrack_tuple_hash *i, 
		  const struct ip_conntrack_helper *me)
{
	if (i->ctrack->helper == me) {
		i->ctrack->helper = NULL;
		/* Get rid of any expected. */
		if (list_inlist(&expect_list, &i->ctrack->expected))
			LIST_DELETE(&expect_list, &i->ctrack->expected);
	}
	return 0;
}

void ip_conntrack_helper_unregister(struct ip_conntrack_helper *me)
{
	unsigned long flags;
	unsigned int i;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	/* Get rid of expecteds, set helpers to NULL. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++)
		LIST_FIND(&ip_conntrack_hash[i], unhelp,
			  struct ip_conntrack_tuple_hash *, me);

	LIST_DELETE(&helpers, me);
	MOD_DEC_USE_COUNT;
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
}

/* Delete all the conntracks. */
void ip_conntrack_cleanup(void)
{
	unsigned long flags;
	unsigned int i;

	LOCK_IRQSAVE(&ip_conntrack_lck, flags);
	/* Mishearing the voices in his head, our hero wonders how
           he's supposed to kill the mall. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		while (!list_empty(&ip_conntrack_hash[i])) {
			ip_conntrack_delete(((struct ip_conntrack_tuple_hash*)
					     ip_conntrack_hash[i].next)
					    ->ctrack);
		}
	}
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
}

void ip_conntrack_init(void)
{
	unsigned int i;
	unsigned long flags;

	printk("ip_conntrack.o v%s\n", NETFILTER_VERSION);
	/* Don't NEED lock here, but good form anyway. */
	LOCK_IRQSAVE(&ip_conntrack_lck, flags);

	/* Sew in builtin protocols. */
	list_append(&protocol_list, &ip_conntrack_protocol_tcp);
	list_append(&protocol_list, &ip_conntrack_protocol_udp);
	list_append(&protocol_list, &ip_conntrack_protocol_icmp);

	ip_conntrack_protocol_tcp_init();

	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++)
		INIT_LIST_HEAD(&ip_conntrack_hash[i]);
	UNLOCK_IRQRESTORE(&ip_conntrack_lck, flags);
}
