/* netfilter device.
 *
 * Rusty Russell (C)1998 -- This code is GPL.
 */
#include <linux/fs.h>
#include <linux/skbuff.h>
#include <linux/if.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/poll.h>
#include <linux/smp_lock.h>
#include <asm/uaccess.h>
#include "device/netfilter_dev.h"

#define NETFILTER_MAJOR 120

#define DEBUG

/* Looking for memcpy?  Could be invoked here if a struct changes. */
#define NEW(type, prio, contents...) 				\
({ 								\
	typeof (type *) _n = kmalloc(sizeof(type), prio); 	\
	if (_n) *_n = ( (type) ## contents);			\
	_n;							\
})

struct nfdev_interest
{
	struct nf_interest base;
	/* Linked list to track of groups of nf_interest */
	struct nfdev_interest *sibling;
};

struct nfdev_private
{
	struct nfdev_interest *interests;

	struct nf_wakeme wake;
	struct nfdev_head head;
	struct sk_buff *current_skb;
};

#define DEVICE_LENGTH(priv) ((priv)->head.pktlen + sizeof(struct nfdev_head))

#ifdef DEBUG
#define NFDEBUG(format, args...)  printk(format , ## args)
#else
#define NFDEBUG(format, args...)
#endif

static loff_t nf_llseek(struct file *file, loff_t offset, int origin)
{
	struct nfdev_private *priv = file->private_data;

	if (!priv->current_skb) {
		NFDEBUG("Pid %u called %s with no packet waiting.\n",
			(unsigned int)current->pid, __FUNCTION__);
		return -EINVAL;
	}

	switch (origin) {
		case 2:
			offset += (int)DEVICE_LENGTH(priv);
			break;
		case 1:
			offset += file->f_pos;
	}

	/* Can seek to end+1, but not past, ie. no holes. */
	if (offset >= 0 && offset <= DEVICE_LENGTH(priv)) {
		file->f_pos = offset;
		return offset;
	}
	NFDEBUG("Pid %u seek to %li invalid.\n",
		(unsigned int)current->pid, (long int)offset);
	return -EINVAL;
}

/* Fill in other fields from priv->current_skb */
static void setheadinfo(struct nfdev_private *priv)
{
	struct net_device *indev, *outdev;
	priv->head.pktlen = priv->current_skb->len;
	nf_getinfo(priv->current_skb, &indev, &outdev, &priv->head.mark);

	if (indev) strcpy(priv->head.iniface, indev->name);
	else priv->head.iniface[0] = '\0';
	if (outdev) strcpy(priv->head.outiface, outdev->name);
	else priv->head.outiface[0] = '\0';
}

static int waitforpacket(struct nfdev_private *priv)
{
	return wait_event_interruptible(priv->wake.sleep, 
					(priv->current_skb
					= skb_dequeue(&priv->wake.skbq)));
}

static ssize_t nf_read(struct file *file, char *buf, size_t count, 
		       loff_t *ppos)
{
	struct nfdev_private *priv = file->private_data;
	size_t size, headsize = 0;

	if (!priv->current_skb) {
		NFDEBUG("Pid %u called %s with no packet waiting.\n",
			(unsigned int)current->pid, __FUNCTION__);
		if (file->f_flags & O_NONBLOCK) 
			return -EAGAIN;

		if (waitforpacket(priv) != 0) return -ERESTARTSYS;
		NFDEBUG("nf_read: Got packet.\n");
		setheadinfo(priv);
	}

	size = UMIN(count, DEVICE_LENGTH(priv) - *ppos);
	if (*ppos < sizeof(struct nfdev_head)) {
		/* Read in part of info structure */
		headsize = UMIN(sizeof(struct nfdev_head), size);
		NFDEBUG("Copying %u head bytes.\n", headsize);
		if (copy_to_user(buf, (char *)&priv->head + *ppos, headsize))
			return -EFAULT;
		size -= headsize;
		*ppos += headsize;
		buf += headsize;
	}

	NFDEBUG("Copying %u data bytes.\n", size);
	if (copy_to_user(buf, 
			 priv->current_skb->data + *ppos - sizeof(struct nfdev_head),
			 size))
		return -EFAULT;

	*ppos += size;
	return (ssize_t)(size + headsize);
}

static ssize_t nf_write(struct file *file, const char *buf,
			size_t count, loff_t *ppos)
{
	struct nfdev_private *priv = file->private_data;

	if (!priv->current_skb) {
		NFDEBUG("Pid %u called %s with no packet waiting.\n",
			(unsigned int)current->pid, __FUNCTION__);
		return -EINVAL;
	}

	if (*ppos < sizeof(struct nfdev_head)) {
		NFDEBUG("Pid %u called %s at offset %lu.\n", current->pid,
			__FUNCTION__, (long int)*ppos);
		return -EINVAL;
	}

	if (*ppos + count > DEVICE_LENGTH(priv)) {
		u_int32_t newlen = *ppos + count - sizeof(struct nfdev_head);

		NFDEBUG("Pid %u called %s over to %lu.\n", current->pid,
			__FUNCTION__, (long int)(*ppos + count));
		/* FIXME: Extend skbuff --RR */
		if (newlen > priv->current_skb->len + skb_tailroom(priv->current_skb)) 
			return -ENOBUFS;

		skb_put(priv->current_skb, newlen - priv->current_skb->len);
		priv->head.pktlen = newlen;
	}

	if (copy_from_user(priv->current_skb->data 
                           + *ppos - sizeof(struct nfdev_head), buf, count))
		return -EFAULT;

	*ppos += count;
	return count;
}

static unsigned int nf_poll(struct file *file, struct poll_table_struct *wait)
{
	struct nfdev_private *priv = file->private_data;
	unsigned int mask = 0;

	/* Already got skb? */
	if (priv->current_skb)
		mask |= POLLOUT | POLLWRNORM | POLLIN | POLLRDNORM;
	else {
		poll_wait(file, &priv->wake.sleep, wait);

		priv->current_skb = skb_dequeue(&priv->wake.skbq);
		if (priv->current_skb) {
			mask |= POLLIN | POLLRDNORM;
			setheadinfo(priv);
		}
	}

	return mask;
}

static int nf_ioctl(struct inode *inode, struct file *file,
		    unsigned int cmd, unsigned long arg)
{
	struct nfdev_private *priv = file->private_data;
	unsigned int proto = MINOR(inode->i_rdev);

	switch (cmd) {
	case NFDIOSVERDICT: {
		struct nfdev_verdict iodata;

		if (!priv->current_skb) {
			NFDEBUG("Pid %u called %s with no packet waiting.\n",
				(unsigned int)current->pid, __FUNCTION__);
			return -EINVAL;
		}

		if (copy_from_user(&iodata, (void *)arg, sizeof(iodata)))
			return -EFAULT;

		if (iodata.pktlen > priv->current_skb->len) {
			NFDEBUG("Pid %u tried to finalize len %u > %u.\n",
				(unsigned int)current->pid, 
				iodata.pktlen, priv->current_skb->len);
			return -EINVAL;
		}
		skb_trim(priv->current_skb, iodata.pktlen);

		nf_reinject(priv->current_skb, iodata.nfmark, iodata.verdict);
		priv->current_skb = NULL;
		file->f_pos = 0;
		return 0;
	}

	case NFDIOSADDCOND: {
		struct nfdev_condition cond;
		struct nfdev_interest *interest;

		if (copy_from_user(&cond, (void *)arg, sizeof(cond)))
			return -EFAULT;

		printk("netfilter_dev: kmalloc()\n");
		/* Add a condition. */
		interest = NEW(struct nfdev_interest, GFP_KERNEL, 
			       { { { NULL, NULL }, proto, cond.hookmask,
				   cond.mark, cond.reason, &priv->wake }, 
					 priv->interests });
		if (!interest)
			return -ENOMEM;

		printk("netfilter_dev: allocated\n");
		priv->interests = interest;

		printk("netfilter_dev: registering interest %p (sibling %p)\n",
		       priv->interests, priv->interests->sibling);

		nf_register_interest(&priv->interests->base);
		printk("netfilter_dev: registered interest %p (next %p)\n",
		       priv->interests, priv->interests->base.list.next);

		return 0;
	}

	case NFDIOSDELCOND: {
		struct nfdev_condition cond;
		struct nfdev_interest **interest, *tmp;

		if (copy_from_user(&cond, (void *)arg, sizeof(cond)))
			return -EFAULT;

		printk("netfilter_dev: looking for condition.\n");

		/* Do we have this condition? */
		for (interest = &priv->interests; 
		     *interest; 
		     interest = &(*interest)->sibling) {
			if ((*interest)->base.hookmask == cond.hookmask
			    && (*interest)->base.mark == cond.mark
			    && (*interest)->base.reason == cond.reason) 
				break;
		}

		if (!*interest)
			return -ENOENT;

		printk("netfilter_dev: found condition %p (sibling %p).\n",
		       *interest, (*interest)->sibling);
		tmp = *interest;
		*interest = (*interest)->sibling;

		/* Delete reason; doesn't clear current skb if any. */
		printk("netfilter_dev: deleting condition %p.\n", tmp);
		nf_unregister_interest(&tmp->base);
		printk("netfilter_dev: deleted condition %p.\n", tmp);
		kfree(tmp);
		printk("netfilter_dev: freed condition %p.\n", tmp);
		return 0;
	}

	default:
		return -EINVAL;
	}
}

/* minor == protocol */
static int nf_open(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);
	struct nfdev_private *priv;

	if (minor >= NPROTO)
		return -ENODEV;

	file->private_data = priv = kmalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	/* We start catching all hooks; use ioctl to refine. */
	priv->interests = NEW(struct nfdev_interest, GFP_KERNEL,
			      { { { NULL, NULL },
				  minor, 0xFFFFFFFF, 0, 0, &priv->wake },
					NULL });
	if (!priv->interests) {
		kfree(priv);
		return -ENOMEM;
	}

	init_waitqueue_head(&priv->wake.sleep);
	skb_queue_head_init(&priv->wake.skbq);
	nf_register_interest(&priv->interests->base);
	priv->current_skb = NULL;

	MOD_INC_USE_COUNT;
	return 0;
}

static int nf_release(struct inode *inode, struct file *file)
{
	struct nfdev_private *priv = file->private_data;

	/* Death implies dishonor. */
	if (priv->current_skb) 
		nf_reinject(priv->current_skb, priv->head.mark, NF_DROP);

	while (priv->interests) {
		struct nfdev_interest *next = priv->interests->sibling;
		nf_unregister_interest(&priv->interests->base);
		kfree(priv->interests);
		priv->interests = next;
	}
	kfree(priv);

	MOD_DEC_USE_COUNT;
	return 0;
}

static struct file_operations netfilter_fops = {
	nf_llseek,
	nf_read,
	nf_write,
	NULL, 		/* readdir */
	nf_poll,
	nf_ioctl,
	NULL,		/* mmap */
	nf_open,
	NULL,		/* flush */
	nf_release,
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL,		/* revalidate */
	NULL,		/* lock */
};

int __init init_netfilter(void)
{
	if (register_chrdev(NETFILTER_MAJOR, "netfilter", &netfilter_fops)) {
		printk(KERN_ERR "netfilter: unable to get major %d\n", 
		       NETFILTER_MAJOR);
		return -EIO;
	}
	return 0;
}

#ifdef MODULE
int init_module(void)
{
	printk(KERN_INFO 
	       "netfilter module v%s: may you live in interesting times.\n",
	       	NETFILTER_VERSION);
	return init_netfilter();
}

void cleanup_module(void)
{
	unregister_chrdev(NETFILTER_MAJOR, "netfilter");
}
#endif
