/* Library which manipulates firewall rules.  Version 0.1. */

/* Architecture of firewall rules is as follows:
 *
 * Chains go INPUT, FORWARD, OUTPUT then user chains.
 * Each user chain starts with an ERROR node.
 * Every chain ends with an unconditional jump: a RETURN for user chains,
 * and a POLICY for built-ins.
 */

/* (C)1999 Paul ``Rusty'' Russell - Placed under the GNU GPL (See
   COPYING for details). */

#include <assert.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "packet-filter/userspace/libiptc/libiptc.h"

#define IP_VERSION	4
#define IP_OFFSET	0x1FFF

static int sockfd = -1;
static void *iptc_fn = NULL;

struct counter_map
{
	enum {
		COUNTER_MAP_NOMAP,
		COUNTER_MAP_NORMAL_MAP,
		COUNTER_MAP_ZEROED
	} maptype;
	unsigned int mappos;
};

struct iptc_handle
{
	/* Have changes been made? */
	int changed;
	/* Number of rules in here reflects original state. */
	struct ipt_getinfo info;
	struct counter_map *counter_map;
	/* Number in here reflects current state. */
	struct ipt_get_entries entries;
};

static
iptc_handle_t alloc_handle(unsigned int num_rules)
{
	size_t len;
	iptc_handle_t h;

	len = sizeof(struct iptc_handle)
		+ num_rules * sizeof(struct ipt_entry)
		+ num_rules * sizeof(struct counter_map);

	if (!(h = malloc(len))) {
		errno = ENOMEM;
		return NULL;
	}	
	h->changed = 0;
	h->entries.num_entries = num_rules;
	h->counter_map = (struct counter_map*)(&h->entries.entries[num_rules]);

	return h;
}

iptc_handle_t
iptc_init(void)
{
	iptc_handle_t h;
	struct ipt_getinfo info;
	unsigned int i;

	iptc_fn = iptc_init;

	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0
	    || setsockopt(sockfd, IPPROTO_IP, IPT_SO_GET_INFO, &info,
			  sizeof(info)) < 0
	    || !(h = alloc_handle(info.num_entries)))
		return NULL;

	h->info = info;

	/* If we do nothing, all counters map onto the same rules. */
	for (i = 0; i < info.num_entries; i++)
		h->counter_map[i]
			= ((struct counter_map){COUNTER_MAP_NORMAL_MAP, i});

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_GET_ENTRIES,
		       &h->entries,
		       sizeof(h->entries)
		       + info.num_entries * sizeof(struct ipt_entry)) < 0) {
		free(h);
		return NULL;
	}
	return h;
}

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

void
dump_entry(struct ipt_entry *e)
{
	size_t i;

	printf("SRC IP: %u.%u.%u.%u/%u.%u.%u.%u\n",
	       IP_PARTS(e->ip.src.s_addr),IP_PARTS(e->ip.smsk.s_addr));
	printf("DST IP: %u.%u.%u.%u/%u.%u.%u.%u\n",
	       IP_PARTS(e->ip.dst.s_addr),IP_PARTS(e->ip.dmsk.s_addr));
	printf("Interface: `%s'/", e->ip.iniface);
	for (i = 0; i < IFNAMSIZ; i++)
		printf("%c", e->ip.iniface_mask[i] ? 'X' : '.');
	printf("to `%s'/", e->ip.outiface);
	for (i = 0; i < IFNAMSIZ; i++)
		printf("%c", e->ip.outiface_mask[i] ? 'X' : '.');
	printf("\nProtocol: %u\n", e->ip.proto);
	if (e->ip.flags & IPT_F_TOS)
		printf("TOS: %u\n", e->ip.tos);
	printf("Flags: %02X\n", e->ip.flags);
	printf("Invflags: %02X\n", e->ip.invflags);
	printf("Counters: %llu packets, %llu bytes\n",
	       e->counters.pcnt, e->counters.bcnt);
	printf("Cache: %08X ", e->nfcache);
	if (e->nfcache & NFC_ALTERED) printf("ALTERED ");
	if (e->nfcache & NFC_UNKNOWN) printf("UNKNOWN ");
	if (e->nfcache & NFC_IP_SRC) printf("IP_SRC ");
	if (e->nfcache & NFC_IP_DST) printf("IP_DST ");
	if (e->nfcache & NFC_IP_IF_IN) printf("IP_IF_IN ");
	if (e->nfcache & NFC_IP_IF_OUT) printf("IP_IF_OUT ");
	if (e->nfcache & NFC_IP_TOS) printf("IP_TOS ");
	if (e->nfcache & NFC_IP_PROTO) printf("IP_PROTO ");
	if (e->nfcache & NFC_IP_OPTIONS) printf("IP_OPTIONS ");
	if (e->nfcache & NFC_IP_TCPFLAGS) printf("IP_TCPFLAGS ");
	if (e->nfcache & NFC_IP_SRC_PT) printf("IP_SRC_PT ");
	if (e->nfcache & NFC_IP_DST_PT) printf("IP_DST_PT ");
	if (e->nfcache & NFC_IP_PROTO_UNKNOWN) printf("IP_PROTO_UNKNOWN ");
	printf("\n");

	printf("Match name: `%s'\n", e->match_name);
	if (strcmp(e->match_name, "tcp") == 0) {
		printf("srcpt=%u-%u dstpt=%u-%u flags=%02X/%02X ",
		       e->matchinfo.tcp.spts[0], e->matchinfo.tcp.spts[1],
		       e->matchinfo.tcp.dpts[0], e->matchinfo.tcp.dpts[1],
		       e->matchinfo.tcp.flg_mask, e->matchinfo.tcp.flg_cmp);
		printf("option=%u invflags=%02X\n",
		       e->matchinfo.tcp.option, e->matchinfo.tcp.invflags);
	} else if (strcmp(e->match_name, "udp") == 0) {
		printf("srcpt=%u-%u dstpt=%u-%u invflags=%02X\n",
		       e->matchinfo.udp.spts[0], e->matchinfo.udp.spts[1],
		       e->matchinfo.udp.dpts[0], e->matchinfo.udp.dpts[1],
		       e->matchinfo.udp.invflags);
	} else if (strcmp(e->match_name, "icmp") == 0) {
		printf("type=%u code=%u-%u invflags=%02X\n",
		       e->matchinfo.icmp.type,
		       e->matchinfo.icmp.code[0], e->matchinfo.icmp.code[1],
		       e->matchinfo.icmp.invflags);
	}

	printf("Target name: `%s'\n", e->target_name);
	if (strcmp(e->target_name, "") == 0) {
		if (e->targinfo.standard.verdict < 0)
			printf("verdict=%s ",
			       e->targinfo.standard.verdict
			       == -NF_ACCEPT-1 ? "NF_ACCEPT"
			       : e->targinfo.standard.verdict
			       == -NF_DROP-1 ? "NF_DROP"
			       : e->targinfo.standard.verdict
			       == -IPT_RETURN-1 ? "RETURN"
			       : "UNKNOWN");
		else
			printf("verdict=%u ",
			    e->targinfo.standard.verdict);
	} else if (strcmp(e->target_name, IPT_ERROR_TARGET) == 0)
		printf("error=`%s'\n", e->targinfo.error);
}

void
dump_entries(const iptc_handle_t handle)
{
	unsigned int i;

	printf("libiptc v%s.  %u entries.\n",
	       NETFILTER_VERSION, handle->entries.num_entries);
	printf("Hooks: in/fwd/out = %u/%u/%u (%u/%u/%u)\n",
	       handle->info.hook_entry[IPT_HOOK_INPUT],
	       handle->info.hook_entry[IPT_HOOK_FORWARD],
	       handle->info.hook_entry[IPT_HOOK_OUTPUT],
	       IPT_HOOK_INPUT, IPT_HOOK_FORWARD, IPT_HOOK_OUTPUT);
	printf("Underflows: in/fwd/out = %u/%u/%u\n",
	       handle->info.underflow[IPT_HOOK_INPUT],
	       handle->info.underflow[IPT_HOOK_FORWARD],
	       handle->info.underflow[IPT_HOOK_OUTPUT]);
	for (i = 0; i < handle->entries.num_entries; i++) {
		printf("Entry %u:\n", i);
		dump_entry(&handle->entries.entries[i]);
		printf("\n");
	}
}

static int
find_label(unsigned int *pos,
	   const char *name,
	   const iptc_handle_t handle)
{
	unsigned int i;
	const struct ipt_entry *entries = handle->entries.entries;

	if (strcmp(name, IPTC_LABEL_INPUT) == 0)
		*pos = handle->info.hook_entry[IPT_HOOK_INPUT];
	else if (strcmp(name, IPTC_LABEL_FORWARD) == 0)
		*pos = handle->info.hook_entry[IPT_HOOK_FORWARD];
	else if (strcmp(name, IPTC_LABEL_OUTPUT) == 0)
		*pos = handle->info.hook_entry[IPT_HOOK_OUTPUT];
	else {
		for (i = 0; i < handle->entries.num_entries; i++) {
			if (strcmp(entries[i].target_name, IPT_ERROR_TARGET)
			    == 0
			    && strcmp(entries[i].targinfo.error, name) == 0) {
				*pos = i+1;
				return 1;
			}
		}
		return 0;
	}
	return 1;
}

/* Returns the position of the final (ie. unconditional) element. */
static unsigned int
get_chain_end(const iptc_handle_t handle, unsigned int pos)
{
	unsigned int i;

	/* Terminate when we meet a error label or a hook entry. */
	for (i = pos+1; i < handle->entries.num_entries; i++) {
		if (i == handle->info.hook_entry[IPT_HOOK_INPUT]
		    || i == handle->info.hook_entry[IPT_HOOK_FORWARD]
		    || i == handle->info.hook_entry[IPT_HOOK_OUTPUT]
		    || strcmp(handle->entries.entries[i].target_name,
			      IPT_ERROR_TARGET) == 0) {
			return i - 1;
		}
	}
	/* SHOULD NEVER HAPPEN */
	fprintf(stderr, "ERROR: Off end (%u) of chain from %u!\n",
		handle->entries.num_entries, pos);
	return handle->entries.num_entries;
}

static int
check_loop(const iptc_handle_t handle)
{
/* FIXME: Breaks on testsuite: RETURN in FORWARD chain. --RR */
#if 0
	const struct ipt_getinfo *info = &handle->info;
	const unsigned int nbuiltin = 3, nb = info->num_entries;
	unsigned int i, exclude_list[nb], stack[nb];

	for(i = 0 ; i < nbuiltin ; i++){
		const unsigned int end
			= get_chain_end(handle,info->hook_entry[i]);
		unsigned int *head = stack, brol = 0;
		*++head = info->hook_entry[i];
		
		while (*head != end) {
			const struct ipt_entry *rule
				= &handle->entries.entries[*head];

			if (!strcmp(rule->target_name, IPT_STANDARD_TARGET)) {
				int verdict = rule->targinfo.standard.verdict;

				if (verdict == -IPT_RETURN -1) {
					(*--head)++;  /* pop & next */
					continue;
				}
				else if (verdict >= 0){
					unsigned int j;
					exclude_list[brol++]= *head;
					for(j = 0; j < brol ; j++){
						if(verdict == exclude_list[j]){
							errno = ELOOP;
							return 0;  
						}
					}
					*(++head) = verdict; /* push */
					continue;
				}
			}
			(*head)++;  /* next */
		}
	}
#endif
	return 1;
}

/* Iterator functions to run through the chains; prev = NULL means
   first chain.  Returns NULL at end. */
const char *
iptc_next_chain(const char *prev, iptc_handle_t *handle)
{
	unsigned int pos;

	if (!prev)
		pos = 0;
	else {
		if (!find_label(&pos, prev, *handle)) {
			errno = ENOENT;
			return NULL;
		}
		pos = get_chain_end(*handle, pos) + 1;
	}

	if (pos == (*handle)->info.hook_entry[IPT_HOOK_INPUT])
		return IPTC_LABEL_INPUT;
	if (pos == (*handle)->info.hook_entry[IPT_HOOK_FORWARD])
		return IPTC_LABEL_FORWARD;
	if (pos == (*handle)->info.hook_entry[IPT_HOOK_OUTPUT])
		return IPTC_LABEL_OUTPUT;
	if (pos == (*handle)->entries.num_entries-1)
		return NULL;

	return (const char *)&(*handle)->entries.entries[pos].targinfo;
}

/* How many rules in this chain? */
unsigned int
iptc_num_rules(const char *chain, iptc_handle_t *handle)
{
	unsigned int pos = 0;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return (unsigned int)-1;
	}
	return get_chain_end(*handle, pos) - pos;
}

/* Get n'th rule in this chain. */
const struct ipt_entry *
iptc_get_rule(const char *chain,
	      unsigned int n,
	      iptc_handle_t *handle)
{
	unsigned int pos = 0;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return NULL;
	}
	return &(*handle)->entries.entries[pos + n];
}

/* Returns a pointer to the target name of this position. */
const char *
iptc_get_target(const char *chain,
		unsigned int n,
		iptc_handle_t *handle)
{
	unsigned int pos = 0;
	struct ipt_standard *act;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return NULL;
	}
	if (strcmp((*handle)->entries.entries[pos + n].target_name,
		   IPT_STANDARD_TARGET) != 0)
		return (*handle)->entries.entries[pos + n].target_name;

	/* Standard target: evaluate */
	act = &(*handle)->entries.entries[pos + n].targinfo.standard;
	if (act->verdict < 0) {
		switch (-act->verdict - 1) {
		case IPT_RETURN:
			return IPTC_LABEL_RETURN;

		case NF_ACCEPT:
			return IPTC_LABEL_ACCEPT;

		case NF_DROP:
			return IPTC_LABEL_DROP;

		default:
			return "*UNKNOWN*";
		}
	}

	if (act->verdict == pos+n+1)
		return "";

	return (const char *)
		&((*handle)->entries.entries[act->verdict - 1].targinfo);
}

/* Is this a built-in chain? */
int 
iptc_builtin(const char *chain, const iptc_handle_t *handle)
{
	return (strcmp(chain, IPTC_LABEL_INPUT) == 0
		|| strcmp(chain, IPTC_LABEL_FORWARD) == 0
		|| strcmp(chain, IPTC_LABEL_OUTPUT) == 0);
}

/* Get the policy of a given built-in chain */
const char *
iptc_get_policy(const char *chain,
		struct ipt_counters *counters,
		iptc_handle_t *handle)
{
	unsigned int start, end;

	if (strcmp(chain, IPTC_LABEL_INPUT) == 0)
		start = (*handle)->info.hook_entry[IPT_HOOK_INPUT];
	else if (strcmp(chain, IPTC_LABEL_FORWARD) == 0)
		start = (*handle)->info.hook_entry[IPT_HOOK_FORWARD];
	else if (strcmp(chain, IPTC_LABEL_OUTPUT) == 0)
		start = (*handle)->info.hook_entry[IPT_HOOK_OUTPUT];
	else return NULL;

	end = get_chain_end(*handle, start);
	*counters = (*handle)->entries.entries[end].counters;

	return iptc_get_target(chain, end-start, handle);
}

/* Adjusts standard verdict jump positions after an insertion/deletion. */
static int
set_verdict(int position, int delta_rules, iptc_handle_t *handle)
{
        struct ipt_get_entries *ptr = &(*handle)->entries;
	unsigned int i;

	for (i = 0; i < ptr->num_entries; i++) {
		/* Trap: insert of fall-through rule.  Don't change
                   x+1 verdict to x+2. */
		if (!strcmp(ptr->entries[i].target_name,IPT_STANDARD_TARGET) 
		    && ptr->entries[i].targinfo.standard.verdict > position
		    && !(i == position && 
			 ptr->entries[i].targinfo.standard.verdict == i+1)) {
			ptr->entries[i].targinfo.standard.verdict
				+= delta_rules;
		}
	}

	(*handle)->changed = 1;
	return 1; 
}

/* If prepend is set, then we are prepending to a chain: if the
 * insertion position is an entry point, keep the entry point. */
static int
insert_rules(unsigned int num_rules, unsigned int position, int prepend,
	     const struct ipt_entry *insert, iptc_handle_t *handle)
{
	iptc_handle_t newh;
	unsigned int i;

	if (position >= (*handle)->entries.num_entries) {
		errno = EINVAL;
		return 0;
	}

	newh = alloc_handle((*handle)->entries.num_entries + num_rules);
	if (!newh)
		return 0;

	/* Fix up entry points. */
	for (i = 0; i < 3; i++) { /* @@ */
		if ((*handle)->info.hook_entry[i] > position)
			newh->info.hook_entry[i]
				= (*handle)->info.hook_entry[i] + num_rules;
		else
			newh->info.hook_entry[i]
				= (*handle)->info.hook_entry[i];
	}

	newh->info.num_entries = (*handle)->info.num_entries;

	/* Copy pre... */
	memcpy(newh->entries.entries, (*handle)->entries.entries,
	       sizeof(newh->entries.entries[0]) * position);
	/* ... Insert new ... */
	memcpy(newh->entries.entries + position, insert,
	       sizeof(newh->entries.entries[0]) * num_rules);
	/* ... copy post */
	memcpy(newh->entries.entries + position + num_rules,
	       (*handle)->entries.entries + position,
	       sizeof(newh->entries.entries[0])
	       * ((*handle)->info.num_entries - position));

	/* Move counter map. */
	/* Copy pre... */
	memcpy(newh->counter_map, (*handle)->counter_map,
	       sizeof(struct counter_map) * position);
	/* ... copy post */
	memcpy(newh->counter_map + position + num_rules,
	       (*handle)->counter_map + position,
	       sizeof(struct counter_map) * ((*handle)->info.num_entries
					     - position));
	/* Set intermediates to no counter copy */
	for (i = 0; i < num_rules; i++)
		newh->counter_map[position+i]
			= ((struct counter_map){ COUNTER_MAP_NOMAP, 0 });

	free(*handle);
	*handle = newh;
	if(!set_verdict(position, num_rules, handle))
		return 0;

	return check_loop(*handle);
}

static int
replace_rule(unsigned int position, const struct ipt_entry *newrule,
	     iptc_handle_t *handle)
{
	if (position > (*handle)->entries.num_entries) {
		errno = EINVAL;
		return 0;
	}

	/* Replace entry. */
	(*handle)->entries.entries[position] = *newrule;
	/* Counter of old rule no longer valid */
	(*handle)->counter_map[position]
		= ((struct counter_map){ COUNTER_MAP_NOMAP, 0 });
	(*handle)->changed = 1;

	return check_loop(*handle);
}

static int
delete_rules(unsigned int position, unsigned int num_rules,
	     iptc_handle_t *handle)
{
	unsigned int i;

	if (position + num_rules > (*handle)->entries.num_entries) {
		errno = EINVAL;
		return 0;
	}

	/* Fix up entry points. */
	for (i = 0; i < 3; i++) {
		if ((*handle)->info.hook_entry[i] > position + num_rules)
			(*handle)->info.hook_entry[i] -= num_rules;
		else if ((*handle)->info.hook_entry[i] > position)
			/* Shouldn't really happen. */
			(*handle)->info.hook_entry[i] = position;
	}

	/* Fix number */
	(*handle)->entries.num_entries -= num_rules;

	/* Move the rules down. */
	memmove(&(*handle)->entries.entries[position],
		&(*handle)->entries.entries[position + num_rules],
		sizeof((*handle)->entries.entries[position])
		* ((*handle)->entries.num_entries - position));

	/* Move the counter map down. */
	memmove(&(*handle)->counter_map[position],
		&(*handle)->counter_map[position + num_rules],
		sizeof(struct counter_map)
		* ((*handle)->entries.num_entries - position));

	return set_verdict(position, -num_rules , handle);
}

static int
map_target(const char *target, struct ipt_entry *rule,
	   unsigned int pos, const char *chain, const iptc_handle_t handle)
{
	/* Always start target off as ALL NUL, for memcmp. */
	memset(rule->target_name, 0, sizeof(rule->target_name));

	/* Maybe it's empty (=> fall through) */
	if (strcmp(target, "") == 0) {
		strcpy(rule->target_name, IPT_STANDARD_TARGET);
		rule->targinfo.standard.verdict = pos+1;
	}
	/* Maybe it's a standard target name... */
	else if (strcmp(target, IPTC_LABEL_ACCEPT) == 0) {
		strcpy(rule->target_name, IPT_STANDARD_TARGET);
		rule->targinfo.standard.verdict = -NF_ACCEPT - 1;
	}
	else if (strcmp(target, IPTC_LABEL_DROP) == 0) {
		strcpy(rule->target_name, IPT_STANDARD_TARGET);
		rule->targinfo.standard.verdict = -NF_DROP - 1;
	}
	else if (strcmp(target, IPTC_LABEL_RETURN) == 0) {
		strcpy(rule->target_name, IPT_STANDARD_TARGET);
		rule->targinfo.standard.verdict = -IPT_RETURN - 1;
	}
	else if (iptc_builtin(target, &handle)) {
		/* Can't jump to builtins. */
		errno = EINVAL;
		return 0;
	} else {
		/* Maybe it's an existing chain name. */
		unsigned int exists;

		if (find_label(&exists, target, handle)) {
			strcpy(rule->target_name, IPT_STANDARD_TARGET);
			rule->targinfo.standard.verdict = exists;
		} else {
			/* Must be a module. */
			if (strlen(target)+1 > sizeof(rule->target_name)) {
				errno = EINVAL;
				return 0;
			}
			strcpy(rule->target_name, target);
		}
	}
	return 1;
}

/* Insert the entry `fw' in chain `chain' into position `rulenum'. */
int
iptc_insert_entry(const ipt_chainlabel chain,
		      struct ipt_entry *fw,
		      const char *target,
		      unsigned int rulenum,
		      iptc_handle_t *handle)
{
	unsigned int pos;

	iptc_fn = iptc_insert_entry;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if ((int)(pos + rulenum) > get_chain_end(*handle, pos)) {
		errno = E2BIG;
		return 0;
	}
	if (!map_target(target, fw, pos + rulenum, chain, *handle))
		return 0;

	return insert_rules(1, pos + rulenum, rulenum == 0, fw, handle);

}

/* Atomically replace rule `rulenum' in `chain' with `fw'. */
int
iptc_replace_entry(const ipt_chainlabel chain,
		       struct ipt_entry *fw,
		       const char *target,
		       unsigned int rulenum,
		       iptc_handle_t *handle)
{
	unsigned int pos;

	iptc_fn = iptc_replace_entry;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if (pos + rulenum >= get_chain_end(*handle, pos)) {
		errno = EINVAL;
		return 0;
	}

	if (!map_target(target, fw, pos + rulenum, chain, *handle))
		return 0;

	return replace_rule(pos + rulenum, fw, handle);
}

/* Append entry `fw' to chain `chain'.  Equivalent to insert with
   rulenum = length of chain. */
int
iptc_append_entry(const ipt_chainlabel chain,
		      struct ipt_entry *fw,
		      const char *target,
		      iptc_handle_t *handle)
{
	unsigned int pos;

	iptc_fn = iptc_append_entry;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	pos = get_chain_end(*handle, pos);

	if (!map_target(target, fw, pos, chain, *handle))
		return 0;

	return insert_rules(1, pos, 0, fw, handle);
}	

/* Delete the first rule in `chain' which matches `fw'. */
int
iptc_delete_entry(const ipt_chainlabel chain,
		      struct ipt_entry *fw,
		      const char *target,
		      iptc_handle_t *handle)
{
	unsigned int pos, lastrule;

	iptc_fn = iptc_delete_entry;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	lastrule = get_chain_end(*handle, pos);

	for (; pos < lastrule; pos++) {
		if (!map_target(target, fw, pos, chain, *handle))
			return 0;

		/* Ensure counters are the same before comparison. */
		fw->counters = (*handle)->entries.entries[pos].counters;
#if 0
		printf("Deleting:\n");
		dump_entry(fw);
#endif
		if (memcmp(&(*handle)->entries.entries[pos], fw, sizeof(*fw))
		    == 0)
			return delete_rules(pos, 1, handle);
	}

	errno = ENOENT;
	return 0;
}	

/* Delete the rule in position `rulenum' in `chain'. */
int
iptc_delete_num_entry(const ipt_chainlabel chain,
			  unsigned int rulenum,
			  iptc_handle_t *handle)
{
	unsigned int pos;

	iptc_fn = iptc_delete_num_entry;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	if (pos + rulenum >= get_chain_end(*handle, pos)) {
		errno = EINVAL;
		return 0;
	}

	return delete_rules(pos + rulenum, 1, handle);
}

/* Check the packet `fw' on chain `chain'.  Returns the verdict, or
   NULL and sets errno. */
const char *
iptc_check_packet(const ipt_chainlabel chain,
			      struct ipt_entry *entry,
			      iptc_handle_t *handle)
{
	errno = ENOSYS;
	return NULL;
}

/* Flushes the entries in the given chain (ie. empties chain). */
int
iptc_flush_entries(const ipt_chainlabel chain,
		       iptc_handle_t *handle)
{
	unsigned int pos;

	iptc_fn = iptc_flush_entries;
	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	return delete_rules(pos, get_chain_end(*handle, pos) - pos, handle);
}

/* Zeroes the counters in a chain. */
int
iptc_zero_entries(const ipt_chainlabel chain,
		      iptc_handle_t *handle)
{
	unsigned int i, end;
	
	if (!find_label(&i, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}
	end = get_chain_end(*handle, i);

	for (; i <= end; i++) {
		if ((*handle)->counter_map[i].maptype ==COUNTER_MAP_NORMAL_MAP)
			(*handle)->counter_map[i].maptype = COUNTER_MAP_ZEROED;
	}
	(*handle)->changed = 1;

	return 1;
}

/* Creates a new chain. */
int
iptc_create_chain(const ipt_chainlabel chain, iptc_handle_t *handle)
{
	struct ipt_entry newchain[2];
	unsigned int pos;

	iptc_fn = iptc_create_chain;

	/* find_label doesn't cover built-in targets: DROP, ACCEPT
           RETURN. */
	if (find_label(&pos, chain, *handle)
	    || strcmp(chain, IPTC_LABEL_DROP) == 0
	    || strcmp(chain, IPTC_LABEL_ACCEPT) == 0
	    || strcmp(chain, IPTC_LABEL_RETURN) == 0) {
		errno = EEXIST;
		return 0;
	}

	if (strlen(chain)+1 > sizeof(ipt_chainlabel)) {
		errno = EINVAL;
		return 0;
	}

	/* To create a chain, create two rules: error node and unconditional
	 * return. */
	memset(newchain, 0, sizeof(newchain));
	strcpy(newchain[0].target_name, IPT_ERROR_TARGET);
	strcpy(newchain[0].targinfo.error, chain);
	newchain[1].targinfo.standard.verdict = -IPT_RETURN - 1;

	return insert_rules(2, (*handle)->entries.num_entries - 1, 0, newchain,
			    handle);
}

/* Get the number of references to this chain. */
int
iptc_get_references(unsigned int *ref, const ipt_chainlabel chain,
			iptc_handle_t *handle)
{
	unsigned int i;
	unsigned int pos;

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}

	*ref = 0;
	for (i = 0; i < (*handle)->entries.num_entries; i++) {
		if (strcmp((*handle)->entries.entries[i].target_name,
			   IPT_STANDARD_TARGET) == 0
		    && ((*handle)->entries.entries[i].targinfo.standard.verdict
			== pos))
			(*ref)++;
	}
	return 1;
}

/* Deletes a chain. */
int
iptc_delete_chain(const ipt_chainlabel chain,
		      iptc_handle_t *handle)
{
	unsigned int pos;
	unsigned int references;

	if (!iptc_get_references(&references, chain, handle))
		return 0;
	
	iptc_fn = iptc_delete_chain;

	if (iptc_builtin(chain, handle)) {
		errno = EINVAL;
		return 0;
	}

	if (references > 0) {
		errno = EMLINK;
		return 0;
	}

	if (!find_label(&pos, chain, *handle)) {
		errno = ENOENT;
		return 0;
	}
	
	if (get_chain_end(*handle, pos) != pos) {
		errno = ENOTEMPTY;
		return 0;
	}

	return delete_rules(pos - 1, 2, handle);
}

/* Renames a chain. */
int iptc_rename_chain(const ipt_chainlabel oldname,
		      const ipt_chainlabel newname,
		      iptc_handle_t *handle)
{
	unsigned int pos;

	/* find_label doesn't cover built-in targets: DROP, ACCEPT
           RETURN. */
	if (find_label(&pos, newname, *handle)
	    || strcmp(newname, IPTC_LABEL_DROP) == 0
	    || strcmp(newname, IPTC_LABEL_ACCEPT) == 0
	    || strcmp(newname, IPTC_LABEL_RETURN) == 0) {
		errno = EEXIST;
		return 0;
	}

	if (!find_label(&pos, oldname, *handle)
	    || iptc_builtin(oldname, handle)) {
		errno = ENOENT;
		return 0;
	}

	if (strlen(newname)+1 > sizeof(ipt_chainlabel)) {
		errno = EINVAL;
		return 0;
	}

	strcpy((*handle)->entries.entries[pos-1].targinfo.error, newname);
	(*handle)->changed = 1;

	return 1;
}

/* Sets the policy on a built-in chain. */
int
iptc_set_policy(const ipt_chainlabel chain,
		    const ipt_chainlabel policy,
		    iptc_handle_t *handle)
{
	unsigned int hook;
	unsigned int pos;

	/* Figure out which chain. */
	if (strcmp(chain, IPTC_LABEL_INPUT) == 0)
		hook = IPT_HOOK_INPUT;
	else if (strcmp(chain, IPTC_LABEL_FORWARD) == 0)
		hook = IPT_HOOK_FORWARD;
	else if (strcmp(chain, IPTC_LABEL_OUTPUT) == 0)
		hook = IPT_HOOK_OUTPUT;
	else {
		errno = EINVAL;
		return 0;
	}

	pos = get_chain_end(*handle, (*handle)->info.hook_entry[hook]);
	if (strcmp(policy, IPTC_LABEL_ACCEPT) == 0)
		(*handle)->entries.entries[pos].targinfo.standard.verdict
			= -NF_ACCEPT -1 ;
	else if (strcmp(policy, IPTC_LABEL_DROP) == 0)
		(*handle)->entries.entries[pos].targinfo.standard.verdict
			= -NF_DROP - 1;
	else {
		errno = EINVAL;
		return 0;
	}
	(*handle)->counter_map[pos]
		= ((struct counter_map){ COUNTER_MAP_NOMAP, 0 });
	(*handle)->changed = 1;

	return 1;
}

/* Without this, on gcc 2.7.2.3, we get:
   libiptc.c: In function `iptc_commit':
   libiptc.c:833: fixed or forbidden register was spilled.
   This may be due to a compiler bug or to impossible asm
   statements or clauses.
*/
void
subtract_counters(struct ipt_counters *answer,
		       const struct ipt_counters *a,
		       const struct ipt_counters *b)
{
	answer->pcnt = a->pcnt - b->pcnt;
	answer->bcnt = a->bcnt - b->bcnt;
}

int
iptc_commit(iptc_handle_t *handle)
{
	/* Replace, then map back the counters. */
	struct ipt_replace *repl;
	struct ipt_counters_info *newcounters;
	unsigned int i;
	size_t counterlen
		= sizeof(struct ipt_counters_info)
		+ sizeof(struct ipt_counters) * (*handle)->entries.num_entries;

	/* Don't commit if nothing changed. */
	if (!(*handle)->changed)
		goto finished;

	repl = malloc(sizeof(*repl)
		      + ((*handle)->entries.num_entries
			 * sizeof(struct ipt_entry)));
	if (!repl) {
		errno = ENOMEM;
		return 0;
	}

	*repl = ((struct ipt_replace)
		 { (*handle)->entries.num_entries,
		   { (*handle)->info.hook_entry[0],
		     (*handle)->info.hook_entry[1],
		     (*handle)->info.hook_entry[2] },
		   /* Underflows are set to last rule in builtins
                      (ie. the policy) */
		   { get_chain_end(*handle, (*handle)->info.hook_entry[0]),
		     get_chain_end(*handle, (*handle)->info.hook_entry[1]),
		     get_chain_end(*handle, (*handle)->info.hook_entry[2]) },
		   (*handle)->info.num_entries,
		   malloc(sizeof(struct ipt_counters)
			  * (*handle)->info.num_entries),
		   { } });

	if (!repl->counters) {
		errno = ENOMEM;
		return 0;
	}
	memcpy(repl->entry_array, (*handle)->entries.entries,
	       (*handle)->entries.num_entries * sizeof(struct ipt_entry));

	/* These are the counters we're going to put back, later. */
	newcounters = malloc(counterlen);
	if (!newcounters) {
		free(repl->counters);
		errno = ENOMEM;
		return 0;
	}

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_REPLACE, repl,
		       sizeof(*repl)
		       + ((*handle)->entries.num_entries
			  * sizeof(struct ipt_entry))) < 0) {
		free(repl->counters);
		free(newcounters);
		return 0;
	}

	/* Put counters back. */
	newcounters->num_counters = (*handle)->entries.num_entries;
	for (i = 0; i < (*handle)->entries.num_entries; i++) {
		unsigned int mappos = (*handle)->counter_map[i].mappos;
		switch ((*handle)->counter_map[i].maptype) {
		case COUNTER_MAP_NOMAP:
			newcounters->counters[i]
				= ((struct ipt_counters){ 0, 0 });
			break;

		case COUNTER_MAP_NORMAL_MAP:
			/* Original read: X.
			 * Atomic read on replacement: X + Y.
			 * Currently in kernel: X.
			 * Want in kernel: X + Y.
			 * => Add in Y
			 * => Add in (replacement read - original read).
			 */
			subtract_counters(&newcounters->counters[i],
					  &repl->counters[mappos],
					  &repl->entry_array[i].counters);
			break;

		case COUNTER_MAP_ZEROED:
			/* Original read: X.
			 * Atomic read on replacement: X + Y.
			 * Currently in kernel: X.
			 * Want in kernel: Y.
			 * => Add in Y - X
			 * => Add in (replacement read - 2*original read).
			 */
			subtract_counters(&newcounters->counters[i],
					  &repl->counters[mappos],
					  &repl->entry_array[i].counters);
			subtract_counters(&newcounters->counters[i],
					  &newcounters->counters[i],
					  &repl->entry_array[i].counters);
			break;
		}
	}

	if (setsockopt(sockfd, IPPROTO_IP, IPT_SO_ADD_COUNTERS, newcounters,
		       counterlen) < 0) {
		free(repl->counters);
		free(newcounters);
		return 0;
	}

	free(repl->counters);
	free(newcounters);

 finished:
	free(*handle);
	*handle = NULL;
	return 1;
}

/* Get raw socket. */
int
iptc_get_raw_socket()
{
	return sockfd;
}

/* Translates errno numbers into more human-readable form than strerror. */
const char *
iptc_strerror(int err)
{
	unsigned int i;
	struct table_struct {
		void *fn;
		int err;
		const char *message;
	} table [] =
	  { { NULL, 0, "Incompatible with this kernel" },
	    { NULL, ENOPROTOOPT, "iptables who? (do you need to insmod?)" },
	    { NULL, ENOSYS, "Will be implemented real soon.  I promise." },
	    { NULL, ENOMEM, "Memory allocation problem" },
	    { iptc_init, EPERM, "Permission denied (you must be root)" },
	    { iptc_delete_chain, ENOTEMPTY, "Chain is not empty" },
	    { iptc_delete_chain, EINVAL, "Can't delete built-in chain" },
	    { iptc_delete_chain, EMLINK, 
	      "Can't delete chain with references left" },
	    { iptc_create_chain, EEXIST, "Chain already exists" },
	    { iptc_insert_entry, E2BIG, "Index of insertion too big" },
	    { iptc_insert_entry, EINVAL, "Target problem" },
	    /* EINVAL for CHECK probably means bad interface. */
	    { iptc_check_packet, EINVAL,
	      "bad arguments (does that interface exist?)" },
	    /* ENOENT for DELETE probably means no matching rule */
	    { iptc_delete_entry, ENOENT,
	      "bad rule (does a matching rule exist in that chain?)" },
	    { NULL, ENOENT, "No extended target/match by that name" }
	  };

	for (i = 0; i < sizeof(table)/sizeof(struct table_struct); i++) {
		if ((!table[i].fn || table[i].fn == iptc_fn)
		    && table[i].err == err)
			return table[i].message;
	}

	return strerror(err);
}
